/*******************************************************************
 Copyright (C) 2006 John Stewart, CRC Canada.
 DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
 See the GNU Library General Public License (file COPYING in the distribution)
 for conditions of use and redistribution.
*********************************************************************/

#include "headers.h"
#include "EAIheaders.h"
#include "jsVRMLClasses.h"

void set_one_ECMAtype (uintptr_t tonode, int toname, int dataType, void *Data, unsigned datalen);
int setMFElementtype (uintptr_t num);

/********************************************************************

getField_ToJavascript.

this sends events to scripts that have eventIns defined.

********************************************************************/
void getField_ToJavascript (int num, int fromoffset) {
	int ignored;
	#ifdef SETFIELDVERBOSE 
		printf ("CRoutes, sending ScriptEventIn to from offset %d type %d\n",
			fromoffset,JSparamnames[fromoffset].type);  
	#endif

	/* set the parameter */
	/* see comments in gatherScriptEventOuts to see exact formats */

	switch (JSparamnames[fromoffset].type) {
	case FIELDTYPE_SFBool:
	case FIELDTYPE_SFFloat:
	case FIELDTYPE_SFTime:
	case FIELDTYPE_SFInt32:
	case FIELDTYPE_SFString:
		setScriptECMAtype(num);
		break;
	case FIELDTYPE_SFColor:
	case FIELDTYPE_SFNode:
	case FIELDTYPE_SFVec2f:
	case FIELDTYPE_SFVec3f:
	case FIELDTYPE_SFRotation:
		setScriptMultiElementtype(num);
		break;
	case FIELDTYPE_MFColor:
	case FIELDTYPE_MFVec3f:
	case FIELDTYPE_MFVec2f:
	case FIELDTYPE_MFFloat:
	case FIELDTYPE_MFTime:
	case FIELDTYPE_MFInt32:
	case FIELDTYPE_MFString:
	case FIELDTYPE_MFNode:
	case FIELDTYPE_MFRotation:
	case FIELDTYPE_SFImage:
		ignored = setMFElementtype(num);
		break;
	default : {
		printf("WARNING: sendScriptEventIn type %s not handled yet\n",
			FIELDTYPES[JSparamnames[fromoffset].type]);
		}
	}
}


/******************************************************************************/

void set_one_ECMAtype (uintptr_t tonode, int toname, int dataType, void *Data, unsigned datalen) {

	char scriptline[100];
	jsval retval;
	jsval newval;
	float fl;
	double dl;
	int il;
	int intval = 0;
	JSContext *cx;
	JSObject *obj, *_sfvec3fObj;

	#ifdef SETFIELDVERBOSE
	printf ("set_one_ECMAtype, to %d namepointer %d, fieldname %s, datatype %d length %d\n",
		tonode,toname,JSparamnames[toname].name,dataType,datalen); 
	#endif

	/* get context and global object for this script */
	cx = (JSContext *) ScriptControl[tonode].cx;
	obj = (JSObject *)ScriptControl[tonode].glob;

	/* set the time for this script */
	SET_JS_TICKTIME


	switch (dataType) {

		case FIELDTYPE_SFFloat:	{
			memcpy ((void *) &fl, Data, datalen);
			newval = DOUBLE_TO_JSVAL(JS_NewDouble(cx,(double)fl));
			break;
		}
		case FIELDTYPE_SFTime:	{
			memcpy ((void *) &dl, Data, datalen);
			newval = DOUBLE_TO_JSVAL(JS_NewDouble(cx,dl));
			break;
		}
		case FIELDTYPE_SFBool:
		case FIELDTYPE_SFInt32: 	{ 
			memcpy ((void *) &il,Data, datalen);
			newval = INT_TO_JSVAL(il);
			break;
		}

		case FIELDTYPE_SFString: {
			struct Uni_String *ms;
			memcpy((void *) &ms,Data, datalen);
			newval = STRING_TO_JSVAL(JS_NewString(cx,ms->strptr,strlen(ms->strptr)));
			break;
		}
		default: {	printf("WARNING: SHOULD NOT BE HERE! %d\n",JSparamnames[toname].type);
		}
	}

	/* get the variable name to hold the incoming value */
	sprintf (scriptline,"__eventIn_Value_%s", JSparamnames[toname].name);
        if (!JS_DefineProperty(cx,obj, scriptline, newval, JS_PropertyStub, JS_PropertyStub, JSPROP_PERMANENT)) {  
                printf( "JS_DefineProperty failed for \"ECMA in\" at %s:%d.\n",__FILE__,__LINE__); 
                return; 
        }

	/* is the function compiled yet? */
	COMPILE_FUNCTION_IF_NEEDED(toname)

	/* and run the function */
	RUN_FUNCTION (toname)
}

/*  setScriptECMAtype called by getField_ToJavascript for 
        case FIELDTYPE_SFBool:
        case FIELDTYPE_SFFloat:
        case FIELDTYPE_SFTime:
        case FIELDTYPE_SFInt32:
        case FIELDTYPE_SFString:
*/

void setScriptECMAtype (uintptr_t num) {
	uintptr_t fn, tn;
	int tptr;
	int len;
	unsigned int to_counter;
	CRnodeStruct *to_ptr = NULL;

	fn = (uintptr_t)(CRoutes[num].fromnode) + (uintptr_t)(CRoutes[num].fnptr);
	len = CRoutes[num].len;

	for (to_counter = 0; to_counter < CRoutes[num].tonode_count; to_counter++) {
		to_ptr = &(CRoutes[num].tonodes[to_counter]);
		tn = (uintptr_t) to_ptr->node;
		tptr = to_ptr->foffset;
		set_one_ECMAtype (tn, tptr, JSparamnames[tptr].type, (void *)fn,len);
	}
}

/* setMFElementtype called by getField_ToJavascript for
         case FIELDTYPE_MFColor:
        case FIELDTYPE_MFVec3f:
        case FIELDTYPE_MFVec2f:
        case FIELDTYPE_MFFloat:
        case FIELDTYPE_MFTime:
        case FIELDTYPE_MFInt32:
        case FIELDTYPE_MFString:
        case FIELDTYPE_MFNode:
        case FIELDTYPE_MFRotation:
        case FIELDTYPE_SFImage:
*/

int setMFElementtype (uintptr_t num) {
	void * fn;
	void * tn;
	uintptr_t tptr, fptr;
	int len;
	unsigned int to_counter;
	CRnodeStruct *to_ptr = NULL;
	char scriptline[20000];
	char sline[100];
	jsval lenval;
	int x;
	int elementlen;
	char *pptr;
	float *fp;
	int *ip;
	double *dp;
	struct Multi_Node *mfp;

	/* for MFStrings we have: */
	char *chptr;
	struct Uni_String  **ptr;

	int indexPointer;

	JSContext *cx;
	JSObject *obj;

	/* for PixelTextures we have: */
	struct X3D_PixelTexture *mePix;
	struct Multi_Int32 image;
	int pixcount;

	#ifdef SETFIELDVERBOSE 
		printf("------------BEGIN setMFElementtype ---------------\n");
	#endif


	fn = CRoutes[num].fromnode;
	fptr = CRoutes[num].fnptr;
	
	/* we can do arithmetic on character pointers; so we have to cast void *
	   to char * here */
	pptr = (char *)fn + fptr;

	len = CRoutes[num].len;

	/* is this from a MFElementType? positive lengths in routing table  == easy memcpy types */
	if (len <= 0) {
		mfp = (struct Multi_Node *) pptr;

		/* check Multimemcpy for C to C routing for this type */
		/* get the number of elements */
		len = mfp->n;  
		pptr = (char *) mfp->p; /* pptr is a char * just for math stuff */
		#ifdef SETFIELDVERBOSE 
		printf ("setMFElementtype, len now %d, from %d\n",len,fn);
		#endif
	} else {
		/* SFImages will have a length of greater than zero */
		/* printf ("setMFElementtype, length is greater than 0 (%d), how can this be?\n",len); */
	}

		

	/* go through all the nodes that this script sends to for this entry in the CRoutes table */
	for (to_counter = 0; to_counter < CRoutes[num].tonode_count; to_counter++) {
		to_ptr = &(CRoutes[num].tonodes[to_counter]);
		tn = to_ptr->node;
		tptr = to_ptr->foffset;
		indexPointer = (uintptr_t) tn; /* tn should be a small int here - it is script # */

		#ifdef SETFIELDVERBOSE 
			printf ("got a script event! index %d type %d\n",
					num, CRoutes[num].direction_flag);
			printf ("\tfrom %#x from ptr %#x\n\tto %#x toptr %#x\n",fn,fptr,tn,tptr);
			printf ("\tfrom %d from ptr %d\n\tto %d toptr %d\n",fn,fptr,tn,tptr);
			printf ("\tdata length %d\n",len);
			printf ("and, sending it to %s as type %d\n",JSparamnames[tptr].name,
					JSparamnames[tptr].type);
		#endif

		/* get context and global object for this script */
		cx = (JSContext *) ScriptControl[indexPointer].cx;
		obj = (JSObject *)ScriptControl[indexPointer].glob;

		/* set the TickTime (possibly again) for this context */
		SET_JS_TICKTIME

		/* make up the name */
		switch (JSparamnames[tptr].type) {
			case FIELDTYPE_MFRotation: {	
				JSObject *newMFObject;
				JSObject *newSFObject;
				SFRotationNative 	*SFRPptr;

				/* create a new MFRotation object... */
				newMFObject = JS_ConstructObject(cx, &MFRotationClass, NULL ,JS_GetParent(cx, obj));
				ADD_ROOT (cx, newMFObject)

				/* define the "length" property for this object */ 
				DEFINE_LENGTH(cx,newMFObject)

				/* fill in private pointer area */
				elementlen = sizeof (float);
				for (x=0; x<len; x++) {
					/* create a new SFRotation object */
					newSFObject = JS_ConstructObject(cx,&SFRotationClass,NULL, newMFObject);
					if ((SFRPptr = (SFRotationNative *)JS_GetPrivate(cx, newSFObject)) == NULL) {
						printf ("failure in getting SF class at %s:%d\n",__FILE__,__LINE__);
						return;
					}

					/* fill the private pointer area */
					fp = (float *)pptr; SFRPptr->v.r[0] = *fp; pptr += elementlen;
					fp = (float *)pptr; SFRPptr->v.r[1] = *fp; pptr += elementlen;
					fp = (float *)pptr; SFRPptr->v.r[2] = *fp; pptr += elementlen;
					fp = (float *)pptr; SFRPptr->v.r[3] = *fp; pptr += elementlen;

					/* put this object into the MF class */
					if (!JS_DefineElement(cx, newMFObject, (jsint) x, OBJECT_TO_JSVAL(newSFObject),
						JS_PropertyStub, JS_PropertyStub, JSPROP_ENUMERATE)) {
							printf("failure in inserting SF class at %s:%d\n",__FILE__,__LINE__);
					}
				}

				/* set the length of this MF */
				SET_LENGTH (cx,newMFObject,len)

				/* set the global variable with this new MF object */
				SET_EVENTIN_VALUE (cx,obj,tptr,newMFObject)
	
				/* run the function */
				COMPILE_FUNCTION_IF_NEEDED(tptr)
				RUN_FUNCTION(tptr)
				break;
			} 

			case FIELDTYPE_MFVec3f: {	
				JSObject *newMFObject;
				JSObject *newSFObject;
				SFVec3fNative 	*SFRPptr;

				/* create a new MFVec3f object... */
				newMFObject = JS_ConstructObject(cx, &MFVec3fClass, NULL ,JS_GetParent(cx, obj));
				ADD_ROOT (cx, newMFObject)

				/* define the "length" property for this object */ 
				DEFINE_LENGTH(cx,newMFObject)

				/* fill in private pointer area */
				elementlen = sizeof (float);
				for (x=0; x<len; x++) {
					/* create a new SFVec3f object */
					newSFObject = JS_ConstructObject(cx,&SFVec3fClass,NULL, newMFObject);
					if ((SFRPptr = (SFVec3fNative *)JS_GetPrivate(cx, newSFObject)) == NULL) {
						printf ("failure in getting SF class at %s:%d\n",__FILE__,__LINE__);
						return;
					}

					/* fill the private pointer area */
					fp = (float *)pptr; SFRPptr->v.c[0] = *fp; pptr += elementlen;
					fp = (float *)pptr; SFRPptr->v.c[1] = *fp; pptr += elementlen;
					fp = (float *)pptr; SFRPptr->v.c[2] = *fp; pptr += elementlen;

					/* put this object into the MF class */
					if (!JS_DefineElement(cx, newMFObject, (jsint) x, OBJECT_TO_JSVAL(newSFObject),
						JS_PropertyStub, JS_PropertyStub, JSPROP_ENUMERATE)) {
							printf("failure in inserting SF class at %s:%d\n",__FILE__,__LINE__);
					}
				}

				/* set the length of this MF */
				SET_LENGTH (cx,newMFObject,len)

				/* set the global variable with this new MF object */
				SET_EVENTIN_VALUE (cx,obj,tptr,newMFObject)
	
				/* run the function */
				COMPILE_FUNCTION_IF_NEEDED(tptr)
				RUN_FUNCTION(tptr)
				break;
			}

			case FIELDTYPE_MFColor: {	
				JSObject *newMFObject;
				JSObject *newSFObject;
				SFColorNative 	*SFRPptr;

				/* create a new MFColor object... */
				newMFObject = JS_ConstructObject(cx, &MFColorClass, NULL ,JS_GetParent(cx, obj));
				ADD_ROOT (cx, newMFObject)

				/* define the "length" property for this object */ 
				DEFINE_LENGTH(cx,newMFObject)

				/* fill in private pointer area */
				elementlen = sizeof (float);
				for (x=0; x<len; x++) {
					/* create a new SFColor object */
					newSFObject = JS_ConstructObject(cx,&SFColorClass,NULL, newMFObject);
					if ((SFRPptr = (SFColorNative *)JS_GetPrivate(cx, newSFObject)) == NULL) {
						printf ("failure in getting SF class at %s:%d\n",__FILE__,__LINE__);
						return;
					}

					/* fill the private pointer area */
					fp = (float *)pptr; SFRPptr->v.c[0] = *fp; pptr += elementlen;
					fp = (float *)pptr; SFRPptr->v.c[1] = *fp; pptr += elementlen;
					fp = (float *)pptr; SFRPptr->v.c[2] = *fp; pptr += elementlen;

					/* put this object into the MF class */
					if (!JS_DefineElement(cx, newMFObject, (jsint) x, OBJECT_TO_JSVAL(newSFObject),
						JS_PropertyStub, JS_PropertyStub, JSPROP_ENUMERATE)) {
							printf("failure in inserting SF class at %s:%d\n",__FILE__,__LINE__);
					}
				}

				/* set the length of this MF */
				SET_LENGTH (cx,newMFObject,len)

				/* set the global variable with this new MF object */
				SET_EVENTIN_VALUE (cx,obj,tptr,newMFObject)
	
				/* run the function */
				COMPILE_FUNCTION_IF_NEEDED(tptr)
				RUN_FUNCTION(tptr)
				break;
			} 

			case FIELDTYPE_MFVec2f: {	
				JSObject *newMFObject;
				JSObject *newSFObject;
				SFVec2fNative 	*SFRPptr;

				/* create a new MFVec2f object... */
				newMFObject = JS_ConstructObject(cx, &MFVec2fClass, NULL ,JS_GetParent(cx, obj));
				ADD_ROOT (cx, newMFObject)

				/* define the "length" property for this object */ 
				DEFINE_LENGTH(cx,newMFObject)

				/* fill in private pointer area */
				elementlen = sizeof (float);
				for (x=0; x<len; x++) {
					/* create a new SFVec2f object */
					newSFObject = JS_ConstructObject(cx,&SFVec2fClass,NULL, newMFObject);
					if ((SFRPptr = (SFVec2fNative *)JS_GetPrivate(cx, newSFObject)) == NULL) {
						printf ("failure in getting SF class at %s:%d\n",__FILE__,__LINE__);
						return;
					}

					/* fill the private pointer area */
					fp = (float *)pptr; SFRPptr->v.c[0] = *fp; pptr += elementlen;
					fp = (float *)pptr; SFRPptr->v.c[1] = *fp; pptr += elementlen;

					/* put this object into the MF class */
					if (!JS_DefineElement(cx, newMFObject, (jsint) x, OBJECT_TO_JSVAL(newSFObject),
						JS_PropertyStub, JS_PropertyStub, JSPROP_ENUMERATE)) {
							printf("failure in inserting SF class at %s:%d\n",__FILE__,__LINE__);
					}
				}

				/* set the length of this MF */
				SET_LENGTH (cx,newMFObject,len)

				/* set the global variable with this new MF object */
				SET_EVENTIN_VALUE (cx,obj,tptr,newMFObject)
	
				/* run the function */
				COMPILE_FUNCTION_IF_NEEDED(tptr)
				RUN_FUNCTION(tptr)
				break;
			}


			case FIELDTYPE_MFFloat: {	
				JSObject *newMFObject;
				jsval newjsval;
				/* create a new MFFloat object... */
				newMFObject = JS_ConstructObject(cx, &MFFloatClass, NULL ,JS_GetParent(cx, obj));
				ADD_ROOT (cx, newMFObject)

				/* define the "length" property for this object */ 
				DEFINE_LENGTH(cx,newMFObject)

				/* fill in private pointer area */
				elementlen = sizeof (float);
				for (x=0; x<len; x++) {
					/* create a new SFFloat object */
					
					fp = (float *)pptr; 
					newjsval = DOUBLE_TO_JSVAL(JS_NewDouble(cx,(double)*fp));
					pptr += elementlen;

					/* put this object into the MF class */
					if (!JS_DefineElement(cx, newMFObject, (jsint) x, OBJECT_TO_JSVAL(newjsval),
						JS_PropertyStub, JS_PropertyStub, JSPROP_ENUMERATE)) {
							printf("failure in inserting SF class at %s:%d\n",__FILE__,__LINE__);
					}
				}

				/* set the length of this MF */
				SET_LENGTH (cx,newMFObject,len)

				/* set the global variable with this new MF object */
				SET_EVENTIN_VALUE (cx,obj,tptr,newMFObject)
	
				/* run the function */
				COMPILE_FUNCTION_IF_NEEDED(tptr)
				RUN_FUNCTION(tptr)
				break;
			}
			case FIELDTYPE_MFTime: {	
				JSObject *newMFObject;
				jsval newjsval;
				/* create a new MFTime object... */
				newMFObject = JS_ConstructObject(cx, &MFTimeClass, NULL ,JS_GetParent(cx, obj));
				ADD_ROOT (cx, newMFObject)

				/* define the "length" property for this object */ 
				DEFINE_LENGTH(cx,newMFObject)

				/* fill in private pointer area */
				elementlen = sizeof (float);
				for (x=0; x<len; x++) {
					/* create a new SFTime object */
					
					fp = (float *)pptr; 
					newjsval = DOUBLE_TO_JSVAL(JS_NewDouble(cx,(double)*fp));
					pptr += elementlen;

					/* put this object into the MF class */
					if (!JS_DefineElement(cx, newMFObject, (jsint) x, OBJECT_TO_JSVAL(newjsval),
						JS_PropertyStub, JS_PropertyStub, JSPROP_ENUMERATE)) {
							printf("failure in inserting SF class at %s:%d\n",__FILE__,__LINE__);
					}
				}

				/* set the length of this MF */
				SET_LENGTH (cx,newMFObject,len)

				/* set the global variable with this new MF object */
				SET_EVENTIN_VALUE (cx,obj,tptr,newMFObject)
	
				/* run the function */
				COMPILE_FUNCTION_IF_NEEDED(tptr)
				RUN_FUNCTION(tptr)
				break;
			}
			case FIELDTYPE_MFInt32: {	
				JSObject *newMFObject;
				jsval newjsval;
				/* create a new MFInt32 object... */
				newMFObject = JS_ConstructObject(cx, &MFInt32Class, NULL ,JS_GetParent(cx, obj));
				ADD_ROOT (cx, newMFObject)

				/* define the "length" property for this object */ 
				DEFINE_LENGTH(cx,newMFObject)

				/* fill in private pointer area */
				elementlen = sizeof (float);
				for (x=0; x<len; x++) {
					/* create a new SFInt32 object */
					
					ip = (int *)pptr; 
					newjsval = INT_TO_JSVAL(ip);
					pptr += elementlen;

					/* put this object into the MF class */
					if (!JS_DefineElement(cx, newMFObject, (jsint) x, newjsval,
						JS_PropertyStub, JS_PropertyStub, JSPROP_ENUMERATE)) {
							printf("failure in inserting SF class at %s:%d\n",__FILE__,__LINE__);
					}
				}

				/* set the length of this MF */
				SET_LENGTH (cx,newMFObject,len)

				/* set the global variable with this new MF object */
				SET_EVENTIN_VALUE (cx,obj,tptr,newMFObject)
	
				/* run the function */
				COMPILE_FUNCTION_IF_NEEDED(tptr)
				RUN_FUNCTION(tptr)
				break;
			}
			case FIELDTYPE_MFString: {	
				JSObject *newMFObject;
				jsval newjsval;
				/* create a new MFString object... */
				newMFObject = JS_ConstructObject(cx, &MFStringClass, NULL ,JS_GetParent(cx, obj));
				ADD_ROOT (cx, newMFObject)

				/* define the "length" property for this object */ 
				DEFINE_LENGTH(cx,newMFObject)

				/* fill in private pointer area */
				for (x=0; x<len; x++) {
					/* create a new SFString object */
					
					chptr = ptr[x]->strptr;
					newjsval = STRING_TO_JSVAL(JS_NewString(cx,chptr,strlen(chptr)));

					/* put this object into the MF class */
					if (!JS_DefineElement(cx, newMFObject, (jsint) x, newjsval,
						JS_PropertyStub, JS_PropertyStub, JSPROP_ENUMERATE)) {
							printf("failure in inserting SF class at %s:%d\n",__FILE__,__LINE__);
					}
				}

				/* set the length of this MF */
				SET_LENGTH (cx,newMFObject,len)

				/* set the global variable with this new MF object */
				SET_EVENTIN_VALUE (cx,obj,tptr,newMFObject)
	
				/* run the function */
				COMPILE_FUNCTION_IF_NEEDED(tptr)
				RUN_FUNCTION(tptr)
				break;
			}
			case FIELDTYPE_MFNode: {	
				JSObject *newMFObject;
				jsval newjsval;
				/* create a new MFNode object... */
				newMFObject = JS_ConstructObject(cx, &MFNodeClass, NULL ,JS_GetParent(cx, obj));
				ADD_ROOT (cx, newMFObject)

				/* define the "length" property for this object */ 
				DEFINE_LENGTH(cx,newMFObject)

				/* fill in private pointer area */
				elementlen = sizeof (float);
				for (x=0; x<len; x++) {
					ip = (int *)pptr; 
					newjsval = INT_TO_JSVAL(ip);
					pptr += elementlen;

					/* put this object into the MF class */
					if (!JS_DefineElement(cx, newMFObject, (jsint) x, newjsval,
						JS_PropertyStub, JS_PropertyStub, JSPROP_ENUMERATE)) {
							printf("failure in inserting SF class at %s:%d\n",__FILE__,__LINE__);
					}
				}

				/* set the length of this MF */
				SET_LENGTH (cx,newMFObject,len)

				/* set the global variable with this new MF object */
				SET_EVENTIN_VALUE (cx,obj,tptr,newMFObject)
	
				/* run the function */
				COMPILE_FUNCTION_IF_NEEDED(tptr)
				RUN_FUNCTION(tptr)
				break;
			}

			case FIELDTYPE_SFImage:	{
				JSObject *newMFObject;
				jsval newjsval;
				/* create a new MFNode object... */
				newMFObject = JS_ConstructObject(cx, &SFImageClass, NULL ,JS_GetParent(cx, obj));
				ADD_ROOT (cx, newMFObject)

				/* define the "length" property for this object */ 
				DEFINE_LENGTH(cx,newMFObject)

				/* fill in private pointer area */
				mePix = (struct X3D_PixelTexture *) fn;
				if (mePix->_nodeType == NODE_PixelTexture) {
					image = mePix->image;
					if (image.n > 2) {
						len = image.n;
						/* copy over the image */
						for (x=0; x<len; x++) {
							newjsval = INT_TO_JSVAL(image.p[x]);
							/* put this object into the MF class */
							if (!JS_DefineElement(cx, newMFObject, (jsint) x, newjsval,
							JS_PropertyStub, JS_PropertyStub, JSPROP_ENUMERATE)) {
								printf("failure in inserting SF class at %s:%d\n",__FILE__,__LINE__);
							}
						}
					} else {
						len = 3;
						for (x=0; x<len; x++) {
							newjsval = INT_TO_JSVAL(0);
							/* put this object into the MF class */
							if (!JS_DefineElement(cx, newMFObject, (jsint) x, newjsval,
							JS_PropertyStub, JS_PropertyStub, JSPROP_ENUMERATE)) {
								printf("failure in inserting SF class at %s:%d\n",__FILE__,__LINE__);
							}
						}
					}

				} else {
					/* make up a "zero" Pixeltexture */
						len = 3;
						for (x=0; x<len; x++) {
							newjsval = INT_TO_JSVAL(0);
							/* put this object into the MF class */
							if (!JS_DefineElement(cx, newMFObject, (jsint) x, newjsval,
							JS_PropertyStub, JS_PropertyStub, JSPROP_ENUMERATE)) {
								printf("failure in inserting SF class at %s:%d\n",__FILE__,__LINE__);
							}
						}
				}

				/* set the length of this MF */
				SET_LENGTH (cx,newMFObject,len)

				/* set the global variable with this new MF object */
				SET_EVENTIN_VALUE (cx,obj,tptr,newMFObject)
	
				/* run the function */
				COMPILE_FUNCTION_IF_NEEDED(tptr)
				RUN_FUNCTION(tptr)

				break;
				} 

			default: {
					printf ("setMFElement, SHOULD NOT DISPLAY THIS\n");
					strcat (scriptline,"(");
				}
		}

	}
	#ifdef SETFIELDVERBOSE 
		printf("------------END setMFElementtype ---------------\n");
	#endif
	return 0; /* return value never checked; #defines expect a return value */
}


void set_EAI_MFElementtype (int num, int offset, unsigned char *pptr, int len) {

    int tn, tptr;
    char scriptline[2000];
    char sline[100];
    jsval retval;
    int x;
    int elementlen;
    float *fp;
    int *ip;
    double *dp;

    JSContext *cx;
    JSObject *obj;

    #ifdef SETFIELDVERBOSE 
	printf("------------BEGIN set_EAI_MFElementtype ---------------\n");
    #endif


    tn   = num;
    tptr = offset;

    #ifdef SETFIELDVERBOSE 
	printf ("got a script event! index %d\n",num);
	printf ("\tto %#x toptr %#x\n",tn,tptr);
	printf ("\tdata length %d\n",len);
	printf ("and, sending it to %s\n",JSparamnames[tptr].name);
    #endif

    /* get context and global object for this script */
    cx = (JSContext *) ScriptControl[tn].cx;
    obj = (JSObject *)ScriptControl[tn].glob;
    SET_JS_TICKTIME

    /* make up the name */
    sprintf (scriptline,"%s(",JSparamnames[tptr].name);
    switch (JSparamnames[tptr].type) {
      case FIELDTYPE_MFVec3f: {
	  strcat (scriptline, "new MFVec3f(");
	  elementlen = sizeof (float) * 3;
	  for (x=0; x<(len/elementlen); x++) {
	      fp = (float *)pptr;
	      sprintf (sline,"%f %f %f",*fp,
		       *(fp+elementlen),
		       *(fp+(elementlen*2)));
	      if (x < ((len/elementlen)-1)) {
		  strcat(sline,",");
	      }
	      pptr += elementlen;
	      strcat (scriptline,sline);
	  }
	  break;
      }
      case FIELDTYPE_MFColor: {
	  strcat (scriptline, "new MFColor(");
	  elementlen = sizeof (float) * 3;
	  for (x=0; x<(len/elementlen); x++) {
	      fp = (float *)pptr;
	      sprintf (sline,"%f %f %f",*fp,
		       *(fp+elementlen),
		       *(fp+(elementlen*2)));
	      if (x < ((len/elementlen)-1)) {
		  strcat(sline,",");
	      }
	      pptr += elementlen;
	      strcat (scriptline,sline);
	  }
	  break;
      }
      case FIELDTYPE_MFFloat: {
	  strcat (scriptline, "new MFFloat(");
	  elementlen = sizeof (float);
	  for (x=0; x<(len/elementlen); x++) {
	      fp = (float *)pptr;
	      sprintf (sline,"%f",*fp);
	      if (x < ((len/elementlen)-1)) {
		  strcat(sline,",");
	      }
	      pptr += elementlen;
	      strcat (scriptline,sline);
	  }

	  break;
      }
      case FIELDTYPE_MFTime:  {
	  strcat (scriptline, "new MFTime(");
	  elementlen = sizeof (double);
	  for (x=0; x<(len/elementlen); x++) {
	      dp = (double *)pptr;
	      sprintf (sline,"%lf",*dp);
	      if (x < ((len/elementlen)-1)) {
		  strcat(sline,",");
	      }
	      pptr += elementlen;
	      strcat (scriptline,sline);
	  }
	  break;
      }
      case FIELDTYPE_MFInt32: {
	  strcat (scriptline, "new MFInt32(");
	  elementlen = sizeof (int);
	  for (x=0; x<(len/elementlen); x++) {
	      ip = (int *)pptr;
	      sprintf (sline,"%d",*ip);
	      if (x < ((len/elementlen)-1)) {
		  strcat(sline,",");
	      }
	      pptr += elementlen;
	      strcat (scriptline,sline);
	  }
	  break;
      }
      case FIELDTYPE_MFString:{
	  strcat (scriptline, "new MFString(");
	  elementlen = sizeof (float);
	  printf ("ScriptAssign, MFString probably broken\n");
	  for (x=0; x<(len/elementlen); x++) {
	      fp = (float *)pptr;
	      sprintf (sline,"%f",*fp);
	      if (x < ((len/elementlen)-1)) {
		  strcat(sline,",");
	      }
	      pptr += elementlen;
	      strcat (scriptline,sline);
	  }
	  break;
      }
      case FIELDTYPE_MFNode:  {
	  strcat (scriptline, "new MFNode(");
	  elementlen = sizeof (int);
	  for (x=0; x<(len/elementlen); x++) {
	      ip = (int *)pptr;
	      sprintf (sline,"%u",*ip);
	      if (x < ((len/elementlen)-1)) {
		  strcat(sline,",");
	      }
	      pptr += elementlen;
	      strcat (scriptline,sline);
	  }
	  break;
      }
      case FIELDTYPE_MFRotation: {	strcat (scriptline, "new MFRotation(");
      elementlen = sizeof (float)*4;
      for (x=0; x<(len/elementlen); x++) {
	  fp = (float *)pptr;
	  sprintf (sline,"%f %f %f %f",*fp,
		   *(fp+elementlen),
		   *(fp+(elementlen*2)),
		   *(fp+(elementlen*3)));
	  sprintf (sline,"%f",*fp);
	  if (x < ((len/elementlen)-1)) {
	      strcat(sline,",");
	  }
	  pptr += elementlen;
	  strcat (scriptline,sline);
      }
      break;
      }
      default: {
	  printf ("setMFElement, SHOULD NOT DISPLAY THIS\n");
	  strcat (scriptline,"(");
      }
    }

    /* convert these values to a jsval type */
    strcat (scriptline,"))");

    #ifdef SETFIELDVERBOSE 
	printf("ScriptLine: %s\n",scriptline);
    #endif

    if (!ActualrunScript(tn,scriptline,&retval))
      printf ("AR failed in setxx\n");

    #ifdef SETFIELDVERBOSE 
	printf("------------END set_EAI_MFElementtype ---------------\n");
    #endif
}


/****************************************************************/
/* sets a SFVec3f and SFColor in a script 			*/
/* sets a SFRotation and SFVec2fin a script 			*/
/*								*/
/* all *Native types have the same structure of the struct -	*/
/* we are just looking for the pointer, thus we can handle	*/
/* multi types here 						*/
/* sets a SFVec3f and SFColor in a script 			*/
/****************************************************************/

/* really do the individual set; used by script routing and EAI sending to a script */
void Set_one_MultiElementtype (uintptr_t tonode, uintptr_t tnfield, void *Data, unsigned dataLen ) {

	char scriptline[100];
	jsval retval;
	SFVec3fNative *_privPtr;

	JSContext *cx;
	JSObject *obj, *_sfvec3fObj;


	/* get context and global object for this script */
	cx = (JSContext *) ScriptControl[tonode].cx;
	obj = (JSObject *)ScriptControl[tonode].glob;

	/* set the time for this script */
	SET_JS_TICKTIME

	/* get the variable name to hold the incoming value */
	sprintf (scriptline,"__eventIn_Value_%s", JSparamnames[tnfield].name);

	#ifdef SETFIELDVERBOSE 
	printf ("Set_one_MultiElementType: script %d line %s\n",tonode, scriptline);
	#endif

	if (!JS_GetProperty(cx,obj,scriptline,&retval))
		printf ("JS_GetProperty failed in Set_one_MultiElementtype.\n");

	if (!JSVAL_IS_OBJECT(retval))
		printf ("Set_one_MultiElementtype - not an object\n");

	_sfvec3fObj = JSVAL_TO_OBJECT(retval);

	if ((_privPtr = (SFVec3fNative *)JS_GetPrivate(cx, _sfvec3fObj)) == NULL)
		printf("JS_GetPrivate failed Set_one_MultiElementtype.\n");

	/* copy over the data from the VRML side into the script variable. */
	memcpy ((void *) &_privPtr->v,Data, dataLen);

	/* is the function compiled yet? */
	COMPILE_FUNCTION_IF_NEEDED(tnfield)

	/* and run the function */
	#ifdef SETFIELDVERBOSE
	printf ("Set_one_MultiElementtype: running script %s\n",scriptline);
	#endif

	RUN_FUNCTION (tnfield)
}

/* setScriptMultiElementtype called by getField_ToJavascript for 
        case FIELDTYPE_SFColor:
        case FIELDTYPE_SFNode:
        case FIELDTYPE_SFVec2f:
        case FIELDTYPE_SFVec3f:
        case FIELDTYPE_SFRotation:
*/

void setScriptMultiElementtype (uintptr_t num) {
	void * fn;
	void * tn;
	uintptr_t tptr, fptr;
	unsigned int len;
	unsigned int to_counter;

	uintptr_t indexPointer;

	CRnodeStruct *to_ptr = NULL;

	JSContext *cx;
	JSObject *obj;

	fn = CRoutes[num].fromnode;
	fptr = CRoutes[num].fnptr;
	len = CRoutes[num].len;

	for (to_counter = 0; to_counter < CRoutes[num].tonode_count; to_counter++) {
		to_ptr = &(CRoutes[num].tonodes[to_counter]);

		/* the to_node should be a script number; it will be a small integer */
		tn = to_ptr->node;
		indexPointer = (uintptr_t) tn;
		tptr = to_ptr->foffset;

		#ifdef SETFIELDVERBOSE 
			printf ("got a script event! index %d type %d\n",
					num, CRoutes[num].direction_flag);
			printf ("\tfrom %#x from ptr %#x\n\tto %#x toptr %#x\n",fn,fptr,indexPointer,tptr);
			printf ("\tdata length %d\n",len);
			printf ("setScriptMultiElementtype here indexPointer %d tptr %d len %d\n",indexPointer, tptr,len);
		#endif

		/* get context and global object for this script */
		cx = (JSContext *) ScriptControl[indexPointer].cx;
		obj = (JSObject *)ScriptControl[indexPointer].glob;
		fn += fptr;
		Set_one_MultiElementtype (indexPointer, tptr, fn, len);
	}
}

/* convert a number in memory to a printable type. Used to send back EVents, or replies to
   the Java client program. */

void EAI_Convert_mem_to_ASCII (int id, char *reptype, int type, char *memptr, char *buf) {

	double dval;
	float fl[4];
	float *fp;
	int *ip;
	int ival;
	int row;			/* MF* counter */
	struct Multi_String *MSptr;	/* MFString pointer */
	struct Multi_Node *MNptr;	/* MFNode pointer */
	struct Multi_Color *MCptr;	/* MFColor pointer */
	char *ptr;			/* used for building up return string */
	struct Uni_String *svptr;
	unsigned char *retSFString;

	int numPerRow;			/* 1, 2, 3 or 4 floats per row of this MF? */
	int i;

	/* used because of endian problems... */
	int *intptr;
	intptr = (int *) memptr;

	switch (type) {
		case FIELDTYPE_SFBool: 	{
			if (eaiverbose) { 
			printf ("EAI_SFBOOL\n");
			}

			if (*intptr == 1) sprintf (buf,"%s\n%f\n%d\nTRUE",reptype,TickTime,id);
			else sprintf (buf,"%s\n%f\n%d\nFALSE",reptype,TickTime,id);
			break;
		}

		case FIELDTYPE_SFTime:	{
			if (eaiverbose) { 
			printf ("EAI_SFTIME\n");
			}
			memcpy(&dval,memptr,sizeof(double));
			sprintf (buf, "%s\n%f\n%d\n%lf",reptype,TickTime,id,dval);
			break;
		}

		case FIELDTYPE_SFInt32:	{
			if (eaiverbose) { 
			printf ("EAI_SFINT32 or EAI_SFNODE\n");
			}
			memcpy(&ival,memptr,sizeof(int));
			sprintf (buf, "%s\n%f\n%d\n%d",reptype,TickTime,id,ival);
			break;
		}

		case FIELDTYPE_SFFloat:	{
			if (eaiverbose) { 
			printf ("EAI_SFFLOAT\n");
			}

			memcpy(fl,memptr,sizeof(float));
			sprintf (buf, "%s\n%f\n%d\n%f",reptype,TickTime,id,fl[0]);
			break;
		}

		case FIELDTYPE_SFVec3f:
		case FIELDTYPE_SFColor:	{
			if (eaiverbose) { 
			printf ("EAI_SFCOLOR or EAI_SFVEC3F\n");
			}
			memcpy(fl,memptr,sizeof(float)*3);
			sprintf (buf, "%s\n%f\n%d\n%f %f %f",reptype,TickTime,id,fl[0],fl[1],fl[2]);
			break;
		}

		case FIELDTYPE_SFVec2f:	{
			if (eaiverbose) { 
			printf ("EAI_SFVEC2F\n");
			}
			memcpy(fl,memptr,sizeof(float)*2);
			sprintf (buf, "%s\n%f\n%d\n%f %f",reptype,TickTime,id,fl[0],fl[1]);
			break;
		}

		case FIELDTYPE_SFColorRGBA:
		case FIELDTYPE_SFRotation:	{
			if (eaiverbose) { 
			printf ("EAI_SFROTATION\n");
			}

			memcpy(fl,memptr,sizeof(float)*4);
			sprintf (buf, "%s\n%f\n%d\n%f %f %f %f",reptype,TickTime,id,fl[0],fl[1],fl[2],fl[3]);
			break;
		}

		case FIELDTYPE_SFImage:
		case FIELDTYPE_SFString:	{
			if (eaiverbose) { 
			printf ("EAI_SFSTRING\n");
			}

			svptr = (struct Uni_String *)memptr;
			retSFString = (unsigned char *)svptr->strptr; 
			sprintf (buf, "%s\n%f\n%d\n\"%s\"",reptype,TickTime,id,retSFString);
			break;
		}

		case FIELDTYPE_MFString:	{
			if (eaiverbose) { 
			printf ("EAI_MFSTRING\n");
			}

			/* make the Multi_String pointer */
			MSptr = (struct Multi_String *) memptr;

			/* printf ("EAI_MFString, there are %d strings\n",(*MSptr).n);*/
			sprintf (buf, "%s\n%f\n%d\n",reptype,TickTime,id);
			ptr = buf + strlen(buf);

			for (row=0; row<(*MSptr).n; row++) {
        	        	/* printf ("String %d is %s\n",row,(*MSptr).p[row]->strptr);*/
				if (strlen ((*MSptr).p[row]->strptr) == 0) {
					sprintf (ptr, "\"XyZZtitndi\" "); /* encode junk for Java side.*/
				} else {
					sprintf (ptr, "\"%s\" ",(*MSptr).p[row]->strptr);
				}
				/* printf ("buf now is %s\n",buf);*/
				ptr = buf + strlen (buf);
			}

			break;
		}

		case FIELDTYPE_MFNode: 	{
			MNptr = (struct Multi_Node *) memptr;

			if (eaiverbose) { 
			printf ("EAI_MFNode, there are %d nodes at %d\n",(*MNptr).n,(intptr_t) memptr);
			}

			sprintf (buf, "%s\n%f\n%d\n",reptype,TickTime,id);
			ptr = buf + strlen(buf);

			for (row=0; row<(*MNptr).n; row++) {
				sprintf (ptr, "%d ",(uintptr_t) (*MNptr).p[row]);
				ptr = buf + strlen (buf);
			}
			break;
		}

		case FIELDTYPE_MFInt32: {
			MCptr = (struct Multi_Color *) memptr;
			if (eaiverbose) { 
				printf ("EAI_MFColor, there are %d nodes at %d\n",(*MCptr).n,(intptr_t) memptr);
			}

			sprintf (buf, "%s\n%f\n%d\n%d \n",reptype,TickTime,id,(*MCptr).n);
			ptr = buf + strlen(buf);

			ip = (int *) (*MCptr).p;
			for (row=0; row<(*MCptr).n; row++) {
				sprintf (ptr, "%d \n",*ip); 
				ip++;
				/* printf ("line %d is %s\n",row,ptr);  */
				ptr = buf + strlen (buf);
			}

			break;
		}

		case FIELDTYPE_MFFloat:
		case FIELDTYPE_MFVec2f:
		case FIELDTYPE_MFVec3f:
		case FIELDTYPE_MFRotation:
		case FIELDTYPE_MFColorRGBA:
		case FIELDTYPE_MFColor: {
			numPerRow=3;
			if (type==FIELDTYPE_MFFloat) {numPerRow=1;}
			else if (type==FIELDTYPE_MFVec2f) {numPerRow=2;}
			else if (type==FIELDTYPE_MFRotation) {numPerRow=4;}
			else if (type==FIELDTYPE_MFColorRGBA) {numPerRow=4;}

			MCptr = (struct Multi_Color *) memptr;
			if (eaiverbose) { 
				printf ("EAI_MFColor, there are %d nodes at %d\n",(*MCptr).n,(intptr_t) memptr);
			}

			sprintf (buf, "%s\n%f\n%d\n%d \n",reptype,TickTime,id,(*MCptr).n);
			ptr = buf + strlen(buf);


			fp = (float *) (*MCptr).p;
			for (row=0; row<(*MCptr).n; row++) {
				for (i=0; i<numPerRow; i++) {
					fl[i] = *fp; fp++;
				}
				switch (numPerRow) {
					case 1:
						sprintf (ptr, "%f \n",fl[0]); break;
					case 2:
						sprintf (ptr, "%f %f \n",fl[0],fl[1]); break;
					case 3:
						sprintf (ptr, "%f %f %f \n",fl[0],fl[1],fl[2]); break;
					case 4:
						sprintf (ptr, "%f %f %f %f \n",fl[0],fl[1],fl[2],fl[3]); break;
				}
				/* printf ("line %d is %s\n",row,ptr); */
				ptr = buf + strlen (buf);
			}

			break;
		}
		default: {
			printf ("EAI, type %c not handled yet\n",type);
		}


/*XXX	case EAI_MFTIME:	{handleptr = &handleEAI_MFTIME_Listener;break;}*/
	}
}

