# Copyright (C) 1998 Tuomas J. Lukka 1999, 2005, 2006 John Stewart CRC Canada.
# DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
# See the GNU Library General Public License (file COPYING in the distribution)
# for conditions of use and redistribution.

# Version of FreeWRL
my $VERSION = "V1.19.8";

# what OS are we running under?
my $OS = `uname`;
my $HAVE_UBUNTU = "NO";
chomp ($pwd = getcwd());

# What compilers do we have?
my $JAVAC = `which javac`;
my $JAR = `which jar`;
if (($JAVAC eq "") || ($JAR eq "")) { 
	print "can not find a java compiler and archiver  - skipping java class making\n";
	$JAVAC = "";
}

# some versions of Linux require the command 
#chcon -t texrel_shlib_t /usr/lib/libFreeWRL*so to be run.

my $CHCON = `which chcon`;
if ($CHCON eq "") {print "chcon command not on system; will NOT run it on install.\n";}
if ($CHCON ne "") {print "chcon command found, will run it on install.\n";}
$CHCON =~ s/\s+$//;

# remove whitespace at beginning and end.
$OS =~ s/\s+$//;
$OS =~ s/^\s+//;

print "\n";
if ($OS eq "Darwin") { print "This install is running on an Apple OS X computer\n";
} elsif ($OS eq "Linux") { print "This install is running on a Linux based computer\n";
} else { print "The uname command returned $OS - dont know what platform - assuming Linux\n"; }

# create the generated code/header files.
system "perl VRMLC.pm";


use Cwd qw(getcwd chdir);


# get the correct.conf file.
if ($OS eq "Darwin") {
	require './vrml.conf.aqua';
} else {
	# assume a Linux computer.
	my $ARCH = `arch`;

	# remove whitespace at beginning and end.
	$ARCH =~ s/\s+$//;
	$ARCH =~ s/^\s+//;

	# find out what the distribution is.
	my ($rpm_prefix,$sourcepackage, $rpmversion, $osname, $specfile) = linux_system_info();
print "specfile $specfile\n";

	# is this an Ubuntu distro?? 
	if (substr($osname,0,6) eq "Ubuntu") {
		$HAVE_UBUNTU = "YES";
		require './vrml.conf.ubuntu';
		print "using \"vrml.conf.ubuntu\" for config variables\n";

	# we can use the generic distros for the architecture.
	} else {
		if ($ARCH eq "x86_64") {
			# 64 bit architecture.
			require './vrml.conf.x86_64';
			printf "using \"vrml.conf.x86_64\" for config variables\n";
		} else {
			# standard 32 bit architecture
			require './vrml.conf';
			printf "using \"vrml.conf\" for config variables\n";
		}
	}

	# add these keys for later use when making distribution (rpm) file
	$VRML_CONFIG{RPMPREFIX} = $rpm_prefix;
	$VRML_CONFIG{SOURCEPACKAGE} = $sourcepackage;
	$VRML_CONFIG{RPMVERSION} = $rpmversion;
	$VRML_CONFIG{OSNAME} = $osname;
	$VRML_CONFIG{SPECFILE} = $specfile;

	# Can we add FreeWRL to the desktop?
	$VRML_CONFIG{DESKTOP} = "NO";	# default value.
	if (-e "/usr/share/applications") {
		if (-e "/usr/share/pixmaps") {
			$VRML_CONFIG{DESKTOP} = "/usr/share";
		}
	}

}

$VRML_CONFIG{VERSION} = $VERSION;

# where are we installing fonts, etc?
if ("" eq $VRML_CONFIG{INSTALL_DIR}) {
	print "do not have an INSTALL_DIR defined, assuming /usr/share\n";
	$VRML_CONFIG{INSTALL_DIR} = "/usr/share";
}

# save this for JS/Makefile.PL
$VRML_CONFIG{PLATFORM} = $OS;

## Handle platform-specific compile options for the Javascript engine
## and others specific to FreeWRL.

my $jsdir = "js1.5/src";
my $jsobjdir;
my $noStackProtector = "";

my @inc;
my @libs;
my @ldflags;
my @jsinc;
my @jslibs;


if ($OS eq  "IRIX64") {
	$jsobjdir = "$jsdir/IRIX6.5";
} elsif ($OS eq "Darwin") {
	$jsobjdir = "$jsdir/Darwin";
} elsif ($OS eq "SunOS") {
	$jsobjdir = "$jsdir/SunOS5.8";
} else { ## linux used by default
	$jsobjdir = "$jsdir/Linux_All";
}

#location of the spidermonkey js engine.
$jsobjdir .= "_OPT.OBJ";
my $libFreeWRLjspath = "./JS/".$jsobjdir."/libFreeWRLjs.so";

#print "the javascript engine should be located at: $libFreeWRLjspath\n";



######################
# Stack Protector flag. Some gcc compiler environments REQUIRE this flag.
# just check to see if it exists.

my $ok;
system ("echo \"int main(){}\" > sp.c");
$ok = system ("gcc -fno-stack-protector sp.c 2>0");
system ("rm sp.c");

# did this -no-stack-protector flag compile ok? If so, lets include it
if ($ok eq 0) { 
	print "adding -fno-stack-protector flag to compiles\n";
	system ("rm a.out");
	$noStackProtector = "-fno-stack-protector";
} else {
	print "this system does not use the -fno-stack-protector flag\n";
}

push @inc, $VRML_CONFIG{FREEWRL_INC}, "-IJS/$jsobjdir", "-IJS/$jsdir", "-ICFuncs -ICFrontEnd ";
push @libs, $VRML_CONFIG{FREEWRL_LIBS}, "-LJS/$jsobjdir", "-lFreeWRLjs";
if ($OS eq "Darwin") {
push @ldflags, $VRML_CONFIG{LDFLAGS}, "-LJS/$jsobjdir";
}

# push build directory, for running standalone

$VRML_CONFIG{FREEWRL_INC} = join(' ', @inc);
$VRML_CONFIG{FREEWRL_LIBS} = join(' ', @libs);

if (defined $VRML_CONFIG{'USE_JAVA_HOME'})
{
	$VRML_CONFIG{JAVA} = $ENV{JAVA_HOME}.'/bin/java';
	$VRML_CONFIG{JAVA_INST} = $ENV{JAVA_HOME};
}

# print out the CFuncs/vrmlconf.c C header file.
open OUT, ">CFuncs/vrmlconf.h" or die("Couldn't write to CFuncs/vrmlconf.h");
print OUT "/* AUTOMATICALLY GENERATED BY TOP-LEVEL MAKEFILE.PL -- DO NOT EDIT */\n\n\n";

# is this building on a Darwin/OSX machine? if so, lets define AQUA.
if ($OS eq "Darwin") {
	print OUT "#define AQUA\n";
}

# lets write some generated #defines.

print OUT "#define BUILDDIR \"$pwd\"\n";
print OUT "#define FWVER \"",$VERSION,"\"\n";
print OUT "#define XMESSAGE \"",$VRML_CONFIG{BIN_DIR},"/FreeWRL_Message\"\n";
print OUT "#define SOUNDSERVERBINARY \"",$VRML_CONFIG{BIN_DIR},"/FreeWRL_SoundServer\"\n";

findAndPrintUtilities();


close OUT;



#######################################################################################
# write the Makefile
#######################################################################################

open MK, ">Makefile" or die("Couldn't write the Makefile");

writeMakefileInitialize();
writeVariables();
writeAllTarget();

writeMakefileLibFreeWRL();
writeMakefileCfrontend();

writeMakefileSoundEngine();
writeMakefileFreeWRLMessage();
write_linux_makefile_javascript();
write_linux_makefile_javaClasses();
writeMakefileBrowserPlugin();
writeMakeDebAndRpm();
writeMakefileDistclean();
writeMakefileFinalTargets();

print "The Makefile has been generated. Type \"make\" and\/or \"make install\" to build (and install)\n\n";

sub writeAllTarget () {
	# should edit this line to suit build
	if ($OS eq "Darwin") {
		print MK "all: Javascript libFreeWRL CFrontEnd";
	} else {
		print MK "all: Javascript libFreeWRL FreeWRL_Message CFrontEnd";
	}
	if ($JAVAC ne "") {print MK " JavaClasses Plugin";}
	if ($VRML_CONFIG{SOUND_ENGINE}) { print MK " SoundEngine"; }
	print MK "\n";
}

sub writeMakefileInitialize() {	
	my $FELIBS = "";

	my $FRONTEND_DIR  = $VRML_CONFIG{BIN_DIR};
	my $LIB_DIR       = $VRML_CONFIG{LIB_DIR};
	my $MYINC         = $VRML_CONFIG{FREEWRL_INC};
	my $MYLIBS 	= $VRML_CONFIG{FREEWRL_LIBS};

        my $MYLDFLAGS     = $VRML_CONFIG{LDFLAGS};

	my $MYDEFINE      = $VRML_CONFIG{FREEWRL_DEFINE};

	if ($OS ne "Darwin") {
		$MYFELIBS =  "-L. ". $MYLIBS . " -lFreeWRLFunc -lpthread";
	} else {
		$MYFELIBS =  "-L. -lpthread ";
	}

	my $MYSELIBS = "-lm ";
	my $MYMESLIBS = "-lXaw ";

	# are we compiling the front end with Motif or GTK2?
	if (index ($VRML_CONFIG{FREEWRL_DEFINE},"HAVE_MOTIF") > 0) {
		 $MYFELIBS = $MYFELIBS . " -lXm ";
	}

	# add on the shared libs, etc. to MYFELIBS for making the freewrl executable.
	# NOTE: order of linking is important; the javascript library should be near
	# the end.
	my $MYFELIBS =  $MYFELIBS." ".$MYLIBS;
	
	
	open ID, ">CFuncs/installdir.h" or die ("Couldn't create installdir.h");
	print ID "#define INSTALLDIR \"$VRML_CONFIG{INSTALL_DIR}/freewrl\"\n";
	close ID;
	
	# do the headers.
	print MK "#Make file for FreeWRL\n";
	print MK "CC = gcc\n";
	print MK "JAVAC = javac\n";
	print MK "INC = $MYINC\n";
	print MK "DEFINE = $noStackProtector $MYDEFINE\n";
	#print MK "FWFLAGS = -c -Wall -O2 -g -pipe -m32 -march=i386 -mtune=pentium4\n";

	# different compiler flags for different architectures
	if ($ARCH eq "x86_64") {
		print MK "FWFLAGS = -c -O2 -g -pipe -m32 -march=i386 -mtune=pentium4\n";
	} else {
		print MK "FWFLAGS = -c -O2 -g -pipe -fno-strict-aliasing -fPIC\n";
	}
	print MK "MYLIBS = $MYLIBS\n";
	print MK "MYFELIBS = $MYFELIBS\n";
	print MK "MYSELIBS = $MYSELIBS\n";
	print MK "MYMESLIBS = $MYMESLIBS\n";
	print MK "FRONTEND_DIR = $FRONTEND_DIR\n";
	print MK "LIB_DIR = $LIB_DIR\n";
	print MK "JAVAC = $JAVAC\n";
	print MK "JAR = $JAR\n";
	print MK "INSTALL_DIR = $VRML_CONFIG{INSTALL_DIR}\n";

	print MK "CFLAGS = \$(FWFLAGS) \$(INC) -IPlugin/netscape/include \$(DEFINE)\n";

	if ($OS eq "Darwin") {
		print MK "MYLDFLAGS = $MYLDFLAGS\n";
	}
}
	
sub writeMakefileFinalTargets() {
	print MK "\n";

	print MK "install:\tall	\n";
	if ($OS ne "Darwin") {
	print MK "\tstrip CFrontEnd/freewrl\n";
	print MK "\tstrip FreeWRL_Msg/FreeWRL_Message\n";
	if ($VRML_CONFIG{SOUND_ENGINE}) { 
		print MK "\tstrip SoundEngine/FreeWRL_SoundServer\n";
	}
	print MK "\tinstall -D CFrontEnd/freewrl \$(DESTDIR)\$(FRONTEND_DIR)/freewrl\n";
	print MK "\tinstall -D FreeWRL_Msg/FreeWRL_Message \$(DESTDIR)\$(FRONTEND_DIR)/FreeWRL_Message\n";

	}
	if ($OS eq "Linux") {
		print MK "\tinstall -D libFreeWRLFunc.so \$(DESTDIR)\$(LIB_DIR)/libFreeWRLFunc.so\n";
		print MK "\tinstall -D $libFreeWRLjspath \$(DESTDIR)\$(LIB_DIR)/libFreeWRLjs.so\n";

		if ($CHCON ne "") {
			print MK "\t-$CHCON -t texrel_shlib_t  \$(DESTDIR)\$(LIB_DIR)/libFreeWRLjs.so\n";
			print MK "\t-$CHCON -t texrel_shlib_t  \$(DESTDIR)\$(LIB_DIR)/libFreeWRLFunc.so\n";
		}
		if ($HAVE_UBUNTU eq "NO") {
			print MK "\t/sbin/ldconfig\n";
		}
	}
	if ($VRML_CONFIG{SOUND_ENGINE}) { 
		print MK "\tinstall -D SoundEngine/FreeWRL_SoundServer \$(DESTDIR)\$(FRONTEND_DIR)/FreeWRL_SoundServer\n";
	}

	
	if (($JAVAC ne "") && ($VRML_CONFIG{NETSCAPE_INST} ne "")) {
		if ($OS ne "Darwin") {
			print MK "\tinstall -D java/classes/vrml.jar \$(DESTDIR)$VRML_CONFIG{NETSCAPE_INST}/vrml.jar\n";
			print MK "\tinstall -D Plugin/npfreewrl.so \$(DESTDIR)$VRML_CONFIG{NETSCAPE_INST}/npfreewrl.so\n";
			if ($CHCON ne "") {
				print MK "\t-$CHCON -t texrel_shlib_t \$(DESTDIR)$VRML_CONFIG{NETSCAPE_INST}/npfreewrl.so\n";
			}
		} else {
			print MK "\tinstall -d  \$(INSTALL_DIR)\/freewrl\n";
			print MK "\tinstall java/classes/vrml.jar \$(INSTALL_DIR)\/freewrl\n";
		}
	}
	if ($OS ne "Darwin") {
	print MK "\tinstall -D Desktop/freewrl.png \$(DESTDIR)\$(INSTALL_DIR)\/pixmaps/Desktop/freewrl.png\n";
	print MK "\tinstall -D Desktop/freewrl.desktop \$(DESTDIR)\$(INSTALL_DIR)\/applications/Desktop/freewrl.desktop\n";
	# we need the 2 installs for the following; the first makes the directory, the second fills it.
	print MK "\tinstall -d fonts \$(DESTDIR)\$(INSTALL_DIR)/freewrl/fonts\n";
	print MK "\tinstall -D fonts/* \$(DESTDIR)\$(INSTALL_DIR)/freewrl/fonts\n";
	} else {
	print MK "\tcp -f freewrl.aquascript /usr/bin/freewrl\n";
	print MK "\tchmod 755 /usr/bin/freewrl\n";
	print MK "\tinstall -d \$(INSTALL_DIR)/freewrl/fonts\n";
	print MK "\tinstall fonts/*.ttf \$(INSTALL_DIR)/freewrl/fonts\n";
	print MK "\tinstall fonts/*.TXT \$(INSTALL_DIR)/freewrl/fonts\n";
	print MK "\tinstall fonts/*.conf \$(INSTALL_DIR)/freewrl/fonts\n";
	}
}

sub writeVariables () {	
	my @CFuncsSourceFiles = split (' ',`ls CFuncs/*c`);
	print MK "\n#CFuncs Make\n";
	print MK "LIBFWSRCS=";
	foreach my $fullname (@CFuncsSourceFiles) { print MK "$fullname "; }
	print MK "\n";
	print MK "LIBFWOBJS=";
	foreach my $fullname (@CFuncsSourceFiles) { $fullname =~ s/.c$/.o/; print MK "$fullname "; }
	print MK "\n\n";
	
	if ($OS ne "Darwin") {
	my @CFrontEndSourceFiles = split (' ',`ls CFrontEnd/*c`);
	print MK "\n#CFrontEnd Make\n";
	print MK "FESRCS=";
	foreach my $fullname (@CFrontEndSourceFiles) { print MK "$fullname "; }
	print MK "\n";
	print MK "FEOBJS=";
	foreach my $fullname (@CFrontEndSourceFiles) { $fullname =~ s/.c$/.o/; print MK "$fullname "; }
	print MK "\n\n";
	}

	my @SoundEngineSourceFiles = split (' ',`ls SoundEngine/*c`);
	print MK "\n#SoundEngine Make\n";
	print MK "SESRCS=";
	foreach my $fullname (@SoundEngineSourceFiles) { print MK "$fullname "; }
	print MK "\n";
	print MK "SEOBJS=";
	foreach my $fullname (@SoundEngineSourceFiles) { $fullname =~ s/.c$/.o/; print MK "$fullname "; }
	print MK "\n\n";

	if ($OS ne  "Darwin") {
	print MK "\n#Console Message Files\n";
	print MK "FWMSRCS=FreeWRL_Msg/FreeWRL_Message.c\n";
	print MK "FWMOBJS=FreeWRL_Msg/FreeWRL_Message.o\n";
	print MK "\n\n";
	}
	
	if ($OS ne "Darwin") {
	print MK "\n#Plugin Files\n";
	print MK "PLSRCS=Plugin/netscape/source/npfreewrl.c Plugin/netscape/source/common/npunix.c\n";
	print MK "PLOBJS=Plugin/netscape/source/npfreewrl.o Plugin/netscape/source/common/npunix.o\n";
	print MK "\n\n";
	}
	
}


sub writeMakefileLibFreeWRL() {
	#############################################################################
	#LIBFREEWRL
	print MK "\$(LIBFWSRCS):\n";
	print MK "\t\$(CC) -c \$*.c\n\n";
	print MK "libFreeWRL:      \$(LIBFWOBJS)\n";
	if ($OS eq "Darwin") { 
	print MK "\t\$(CC) -dynamiclib \$(MYFELIBS) \$(MYLDFLAGS) -o libFreeWRLFunc.dylib  \$(MYLIBS) \$(LIBFWOBJS)\n\n";
	} else {
	print MK "\t\$(CC) -shared -o libFreeWRLFunc.so \$(LIBFWOBJS)\n\n";
	}
}


sub writeMakefileCfrontend() {
	#############################################################################
	#CFRONTEND
	if ($OS ne "Darwin") {
	print MK "\n# make the Unix front end\n";
	print MK "\$(FESRCS):\n";
	print MK "\t\$(CC) -c \$*.c\n\n";
	print MK "CFrontEnd:      \$(FEOBJS)\n";
	print MK "\t\$(CC) \$(FEOBJS) \$(MYFELIBS) -o CFrontEnd/freewrl\n\n";
	}
}

sub writeMakefileSoundEngine() {
	#############################################################################
	# SoundEngine
	# make the sound engine, console message, and frontend if NOT OSX.

	if ($VRML_CONFIG{SOUND_ENGINE}) {
		print MK "\n# make the Unix sound engine \n";
		print MK "\$(SESRCS):\n";
		print MK "\t\$(CC) -c \$*.c\n\n";
		print MK "SoundEngine:      \$(SEOBJS)\n";
		print MK "\t\$(CC)  -o SoundEngine/FreeWRL_SoundServer \$(MYLIBS) \$(SEOBJS) \$(MYSELIBS)\n\n";
	}
}


sub writeMakefileFreeWRLMessage() {
	#############################################################################
	# FreeWRL_Message 
	# make a program that displays a message if NOT OSX and if we are not using Motif.


	if ($OS ne "Darwin") { 
	print MK "\n# make the Unix Console Message \n";
	print MK "\$(FWMSRCS):\n";
	print MK "\t\$(CC) -c \$*.c\n\n";
	print MK "FreeWRL_Message:      \$(FWMOBJS)\n";
	print MK "\t\$(CC) FreeWRL_Msg/FreeWRL_Message.o  -o FreeWRL_Msg/FreeWRL_Message \$(MYLIBS) \$(MYMESLIBS) \n\n";
	}
}

sub write_linux_makefile_javascript() {
	#############################################################################
	## make the Javascript libraries if needed
    
	print MK "Javascript:\n";
	print MK "\tcd JS/$jsdir; make -f Makefile.ref BUILD_OPT=1 STACKPROT=$noStackProtector; cd $pwd\n";
}



sub write_linux_makefile_javaClasses() {
	if ($JAVAC ne "" ) {
		my @JavaSourceFiles = split (' ',`find . -name "*java" | grep "java\/classes"`);
		foreach my $fullname (@JavaSourceFiles) {
			# remove the Java from the front, and the .c from the end.
			my $nm = $fullname;
			$nm =~ s/^\.\///; $nm =~ s/.java$/.class/;
	
			print MK "$nm:\n	\$(JAVAC) -classpath java/classes $fullname\n";
		}
		
		print MK "JavaClasses:\t";
		foreach my $fullname (@JavaSourceFiles) {
			# remove the Java from the front, and the .c from the end.
			my $nm = $fullname;
			$nm =~ s/^\.\///; $nm =~ s/.java$/.class/;
	
			print MK "$nm ";
		}
		print MK "\n";
	
		# make the jar file here
		print MK "	\$(JAR) -cf java/classes/vrml.jar $@ `find java/classes/vrml -name *.class`\n";
	}
}
sub writeMakefileBrowserPlugin() {
	if ($OS ne "Darwin") {
	if ($VRML_CONFIG{NETSCAPE_INST}) {
		print MK "\n# make the Unix front end\n";
		print MK "\$(PLSRCS):\n";
		print MK "\t\$(CC) -c \$*.c\n\n";
		print MK "Plugin:      \$(PLOBJS)\n";
		print MK "\t\$(CC)  -shared -o Plugin/npfreewrl.so \$(PLOBJS)\n\n";
	}
	}
}

sub writeMakeDebAndRpm () {
	if ($OS ne "Darwin") {
		$specfile = $VRML_CONFIG{SPECFILE};
		$rpm_prefix = $VRML_CONFIG{RPMPREFIX};
		$sourcepackage = $VRML_CONFIG{SOURCEPACKAGE};
		$rpmversion = $VRML_CONFIG{RPMVERSION}; 
		$osname = $VRML_CONFIG{OSNAME};

		print MK "deb::\n";
		print MK "\tsudo debuild -S\n";
		print MK "\tsudo pbuilder build ../*.dsc\n";
		print MK "rpm::\n";
		print MK "\t# make the gzip file of the dist.\n";
		print MK "\ttar -cf $rpm_prefix/SOURCES/$sourcepackage ..\/$srcdir\n";
		print MK "\trm -f $rpm_prefix/SOURCES/$sourcepackage.gz\n";
		print MK "\tgzip $rpm_prefix/SOURCES/$sourcepackage\n\n";
		print MK "\t#make the spec file.\n";
		print MK "\techo \"Summary: A vrml viewer\" > $specfile\n";
		print MK "\techo \"Name: freewrl\" >> $specfile\n";
		print MK "\techo \"Version: $rpmversion\" >> $specfile\n";
		print MK "\techo \"Release: $osname\" >> $specfile\n";
		#print MK "\techo \"Copyright: GPL\" >> $specfile\n";
		print MK "\techo \"License: GPL\" >> $specfile\n";
		print MK "\techo \"Group: Applications\/Multimedia\" >> $specfile\n";
		print MK "\techo \"Source0: $sourcepackage.gz\" >> $specfile\n";
		print MK "\techo \"URL: http:\/\/www.crc.ca\/FreeWRL\" >> $specfile\n";
		print MK "\techo \"Distribution: Linux\" >> $specfile\n";
		print MK "\techo \"Vendor: CRC Canada\" >> $specfile\n";
		print MK "\techo \"Packager: Autogenerated, John Stewart, <freewrl-06@rogers.com>\" >> $specfile\n";
		print MK "\techo \"Provides: freewrl\" >> $specfile\n";
		print MK "\techo \"%define __check_files %{nil} \" >> $specfile\n";
		print MK "\techo \"%description\" >> $specfile\n";
		print MK "\techo \"FreeWRL is an Open Source, cross platform VRML2/X3D compliant browser, with script and EAI support.\" >> $specfile\n";
		print MK "\techo \"The primary platform is Linux.\" >> $specfile\n";
		print MK "\techo \"%prep\" >> $specfile\n";
		# split the zcat into 2 lines because we need the $RPM_SOURCE_DIR in the rpm spec file...
		print MK "\techo -n \"zcat \$\$\" >> $specfile\n";
		print MK "\techo \"RPM_SOURCE_DIR\/$sourcepackage.gz | tar -xvf -\" >> $specfile\n";
		print MK "\techo \"%setup -q\" >> $specfile\n";
		print MK "\techo \"%build\" >> $specfile\n";
		print MK "\techo \"perl Makefile.PL\" >> $specfile\n";
		print MK "\techo \"%install\" >> $specfile\n";
		print MK "\techo \"make install\" >> $specfile\n";
		print MK "\techo \"%post\" >> $specfile\n";
		print MK "\techo \"\/sbin\/ldconfig\" >> $specfile\n";
		print MK "\techo \"%clean\" >> $specfile\n";
		print MK "\techo \"%files\" >> $specfile\n";
		print MK "\techo \"%defattr(-,root,root)\" >> $specfile\n";
		print MK "\techo \"\$(DESTDIR)\$(LIB_DIR)/libFreeWRLFunc.so\" >> $specfile\n";
		print MK "\techo \"\$(DESTDIR)\$(LIB_DIR)/libFreeWRLjs.so\" >> $specfile\n";
		print MK "\techo \"\$(DESTDIR)\$(FRONTEND_DIR)/FreeWRL_Message\" >> $specfile\n";
		print MK "\techo \"\$(DESTDIR)\$(FRONTEND_DIR)/freewrl\" >> $specfile\n";
		print MK "\techo \"\$(DESTDIR)\$(FRONTEND_DIR)/FreeWRL_SoundServer\" >> $specfile\n";
                print MK "\techo \"\$(DESTDIR)$VRML_CONFIG{NETSCAPE_INST}/vrml.jar\" >> $specfile\n";
                print MK "\techo \"\$(DESTDIR)$VRML_CONFIG{NETSCAPE_INST}/npfreewrl.so\" >> $specfile\n";
                print MK "\techo \"\$(DESTDIR)\$(INSTALL_DIR)\/pixmaps/Desktop/freewrl.png\" >> $specfile\n";
                print MK "\techo \"\$(DESTDIR)\$(INSTALL_DIR)\/applications/Desktop/freewrl.desktop\" >> $specfile\n";
                print MK "\techo \"\$(DESTDIR)\$(INSTALL_DIR)/freewrl/fonts\" >> $specfile\n";
	}
}


sub writeMakefileDistclean() {
	#############################################################################
	# Misc targets
	print MK "\nclean:\n";
	print MK "	rm -f CFuncs/*o\n";
	print MK "	rm -f libFreeWRLFunc.so\n";
	print MK "	rm -f FreeWRL_Msg/*o\n";
	print MK "	rm -f FreeWRL_Msg/FreeWRL_Message\n";
	print MK "	rm -f CFrontEnd/*o\n";
	print MK "	rm -f CFrontEnd/freewrl\n";
	print MK "	rm -f freewrl\n";
	if ($VRML_CONFIG{SOUND_ENGINE}) {
		print MK "	rm -f SoundEngine/*o\n";
		print MK "	rm -f SoundEngine/FreeWRL_SoundServer\n";
	}
	print MK "	rm -f java/classes/vrml/external/field/*class\n";
	print MK "	rm -f java/classes/vrml/field/*class\n";
	print MK "	rm -f java/classes/vrml/external/node/*class\n";
	print MK "	rm -f java/classes/vrml/external/FreeWRLEAI/*class\n";
	print MK "	rm -f java/classes/vrml/external/*class\n";
	print MK "	rm -f java/classes/vrml/external/exception/*class\n";
	print MK "	rm -f java/classes/vrml/node/*class\n";
	print MK "	rm -f java/classes/vrml/*class\n";
	print MK "	rm -f java/classes/vrml/*class\n";
	print MK "	rm -f java/classes/sai/*class\n";
	print MK "	rm -f java/classes/sai/eai/*class\n";
	print MK "	rm -f java/classes/org/web3d/x3d/sai/*class\n";
	print MK "	rm -f java/classes/vrml.jar\n";
	print MK "	rm -rf JS/$jsobjdir\n";

	if ($OS eq "Darwin") {
		print MK "	rm -f /usr/local/lib/libFreeWRLFunc.dylib\n";
		print MK "	rm -f /usr/local/lib/libFreeWRLjs.dylib\n";
	}

	# remove other JS directories
	$jsobjdir =~s/src/src\/editline/;
	print MK "	rm -rf JS/$jsobjdir\n";
	$jsobjdir =~ s/editline/fdlibm/;
	print MK "	rm -rf JS/$jsobjdir\n";

	print MK "\n#remove installed images\n";
	print MK "distclean:	clean\n";
	print MK "\trm -f \$(DESTDIR)\$(FRONTEND_DIR)/freewrl\n";
	print MK "\trm -f \$(DESTDIR)\$(FRONTEND_DIR)/FreeWRL_Message\n";
	
	if ($OS eq "Linux") {
		print MK "\trm -f \$(DESTDIR)\$(LIB_DIR)/libFreeWRLFunc.so\n";
		print MK "\trm -f \$(DESTDIR)\$(LIB_DIR)/libFreeWRLjs.so\n";
		if ($HAVE_UBUNTU eq "NO") {
			print MK "\t/sbin/ldconfig\n";
		}
	}
	if ($VRML_CONFIG{SOUND_ENGINE}) { 
		print MK "\trm -f \$(DESTDIR)\$(FRONTEND_DIR)/FreeWRL_SoundServer\n";
	}

	if (($JAVAC ne "") && ($VRML_CONFIG{NETSCAPE_INST} ne "")) {
	if ($OS ne "Darwin") {
		print MK "\trm -f \$(DESTDIR)$VRML_CONFIG{NETSCAPE_INST}/vrml.jar\n";
		print MK "\trm -f \$(DESTDIR)$VRML_CONFIG{NETSCAPE_INST}/npfreewrl.so\n";
		print MK "\trm -f Plugin/npfreewrl.so\n";
		print MK "\trm -f Plugin/netscape/source/npfreewrl.o\n";
		print MK "\trm -f Plugin/netscape/source/common/npunix.o\n";
	}
	}
	if ($OS ne "Darwin") {
		print MK "\trm -f \$(DESTDIR)\$(INSTALL_DIR)\/pixmaps/Desktop/freewrl.png\n";
		print MK "\trm -f \$(DESTDIR)\$(INSTALL_DIR)\/applications/Desktop/freewrl.desktop\n";
	}
	print MK "\trm -rf \$(DESTDIR)\$(INSTALL_DIR)/freewrl/fonts\n";
	print MK "\trm -rf \$(DESTDIR)\$(INSTALL_DIR)/freewrl/x3d\n";
}



###################################################################################
# determine on what system we are running, and if we can make a package for it for
# redistribution
sub linux_system_info() {
	
	my $rpm_prefix;
	my $osname;
	my $mightBeUbuntu = FALSE;
	
	# determine where the package should be built.
	# is this a Red Hat, Fedora package?
	#rpm_prefix = "/tmp";
	if (-e '/usr/src/redhat') {
		$rpm_prefix = "/usr/src/redhat";
	
	# is this SuSE/Novell linux??
	} elsif (-e '/usr/src/packages') {
		$rpm_prefix = "/usr/src/packages";

	# yikes - do not know...
	} else { $rpm_prefix = "/tmp"; }
		
	# what OS version is this built on? each Linux release has a different
	# file name here - might be /etc/SuSE-release. 

	my $distrib = `lsb_release --id | cut -f2`;
	$distrib =~ s/\b(\w)/\U$1/g;
	# remove trailing cr
	$distrib =~s/\n.*//;

	my $ver = `lsb_release --release | cut -f2`;
	# remove trailing cr
	$ver =~s/\n.*//;

	$osname = $distrib.$ver;
	
	if ($osname eq "") {
		print "could not determine how to make redistributable packages on this system \n";
		print "so the make rpm target will not work - oh well!\n";
		$osname = "unknown";
	}
	#print "osname is: $osname\n";
	
	
	# sourcepackage is a file created by "packagemaker" - it contains the version, etc.
	my $sourcepackage = `cat sourcepackage`;
	# remove whitespace at beginning and end of this package name (check packagemaker in CVS tree)
	$sourcepackage =~ s/\s+$//;
	$sourcepackage =~ s/^\s+//;
	my $srcdir = substr($sourcepackage, 0, rindex ($sourcepackage,".tar"));
	my $rpmversion = substr ($srcdir,index($srcdir,"-")+1,100);
	#print "srcdir $srcdirrpmversion $rpmversion sourcepackage $sourcepackage\n";
	
	# make spec file name.
	my $specfile = "$rpm_prefix/SPECS/".$srcdir.".".$osname.".spec";
	#print "spec file is $specfile\n";
	
	return ($rpm_prefix,$sourcepackage, $rpmversion, $osname, $specfile);
}

sub findAndPrintUtilities() {
	# lets just dump out some standard defines, from the vrml.conf file.
	my @dumpToVRMLCONF = ( "BROWSER", "SOUNDCONV", "UNZIP", "WGET", "CONVERT", "COPIER", "JAVA");

	my $val;

	foreach (@dumpToVRMLCONF) {

		if (defined $VRML_CONFIG{$_}) {
			if (!(-e $VRML_CONFIG{$_})) {
				print "WARNING: Check your specific vrml.conf file; can not find system executable ".
					$VRML_CONFIG{$_}." for identifier $_\n";
			}
			$val = $VRML_CONFIG{$_};
 		} else {
			# lets see if we can find it....
			$val = tryToFind($_);
	 	}
		print OUT "#define\t$_\t\t\"$val\"\n";
	}
}

sub tryToFind() {
	my ($tofind) = @_;
	my $val = "ISNOTDEFINED";
	my $try = "";
	
	if ("BROWSER" eq $tofind) {
		# more than 1 possibility - remove whitespace after each try
		$try = `which firefox`; $try =~ s/^\s+//; $try =~ s/\s+$//;
		if ("" eq $try) { $try = `which safari`; $try =~ s/^\s+//; $try =~ s/\s+$//; }
		if ("" eq $try) { $try = `which mozilla`; $try =~ s/^\s+//; $try =~ s/\s+$//; }
		if ("" eq $try) { $try = `which netscape`; $try =~ s/^\s+//; $try =~ s/\s+$//; }
		if ("" eq $try) { $try = `which opera`; $try =~ s/^\s+//; $try =~ s/\s+$//; }
		if ("" eq $try) { $try = `which firefox`; $try =~ s/^\s+//; $try =~ s/\s+$//; }
		if ("" eq $try) { $try = `which seamonkey`; $try =~ s/^\s+//; $try =~ s/\s+$//; }
	} elsif ("SOUNDCONV" eq $tofind) {
		$try = `which sox`; $try =~ s/^\s+//; $try =~ s/\s+$//;
	} elsif ("UNZIP" eq $tofind) {
		$try = `which gunzip`; $try =~ s/^\s+//; $try =~ s/\s+$//;
	} elsif ("WGET" eq $tofind) {
		# more than 1 possibility - remove whitespace after each try
		$try = `which wget`; $try =~ s/^\s+//; $try =~ s/\s+$//;
		if ("" eq $try) {$try = `which curl`; $try =~ s/^\s+//; $try =~ s/\s+$//;}
	} elsif ("CONVERT" eq $tofind) {
		$try = `which convert`; $try =~ s/^\s+//; $try =~ s/\s+$//;
	} elsif ("COPIER" eq $tofind) {
		$try = `which cp`; $try =~ s/^\s+//; $try =~ s/\s+$//;
	} elsif ("JAVA" eq $tofind) {
		$try = `which java`; $try =~ s/^\s+//; $try =~ s/\s+$//;
	}

	if ("" ne $try) {
		$val = $try; # found it, hopefully
	} else {
		print "\tWARNING: can not find  $tofind in the conf file, or on the system"; print ": returning: $val\n";		
	}

	return $val;
}
